#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#define PI 3.14159265f

#include <png/png.h>
#include "ptc/tinyptc.h"
#include "playmu/playmu.h"

#define WIDTH 640
#define HEIGHT 480
#define SIZE WIDTH*HEIGHT

#define rnd(n) (rand()%(n+1))

static DWORD ptc_screen[SIZE];
DWORD screen[320*240];

struct Sprite
{
	int width;
	int height;
	DWORD* data;
};

struct Box
{
	int x;
	int y;
	int w;
	int h;
};

struct Entoid
{
	bool exists;
	int type;
	int timer;
	int x;
	int y;
	Sprite* sprite;
	bool flipped;
	int layer;
	// local box
	Box box;
	// global (collision) box
	Box cbox;
	// current messages
	int msg[20];
	int msgcount;
	// new messages
	int msgn[20];
	int msgncount;
	// local data (type specific)
	void* data;
	// update function
	bool (*TickFunc)(Entoid&);
};

struct Platform
{
	bool exists;
	int x;
	int y;
	int w;
	int h;
	Sprite* sprite;
};

Platform platforms[1024];
int maxplatforms=0;

Entoid entoids[1024];
int maxentoids=0;

Sprite spr_room[10];
Sprite spr_gem[5];
Sprite spr_spark[3];
Sprite spr_calvin[3];
Sprite spr_drone[2];
Sprite spr_enemy[6];
Sprite spr_border;
Sprite spr_bg1, spr_bg2;
Sprite spr_win, spr_lose;

Sprite spr_platform[20];
int numptypes=0;

playmu_sound sfx_jump, sfx_gem, sfx_win, sfx_lose, sfx_drone, sfx_splash, sfx_splat, sfx_build;

int scroll_x=0, scroll_y=0;

int global_win=0;
int global_lose=0;
int global_score=0;


// garbage collection (well, automatic cleanup)
#define GCNEW 0
#define GCNEWARRAY 1
#define GCMALLOC 2

void* gcptr[200];
int gcmethod[200];
int numgc=0;

void GCAdd(void* ptr, int method)
{
	gcptr[numgc]=ptr;
	gcmethod[numgc]=method;
	numgc++;
}

void GCFree()
{
	for(int i=0;i<numgc;i++)
	{
		switch(gcmethod[i])
		{
		case GCNEW:
			delete gcptr[i];
			break;
		case GCNEWARRAY:
			delete[] gcptr[i];
			break;
		case GCMALLOC:
			free(gcptr[i]);
			break;
		}
	}
}


#include "gfx.h"

#include "mechanics.h"

void AddEntoid(void (*Spawn)(Entoid&), int x, int y)
{
	for(int i=0;i<maxentoids;i++)
		if(!entoids[i].exists)
			break;
	Entoid& e=entoids[i];
	e.exists=true;
	e.type=-1;
	e.x=x;
	e.y=y;
	e.msgcount=0;
	e.msgncount=0;
	e.timer=0;
	e.flipped=false;
	e.layer=0;
	Spawn(e);
	e.cbox.x=-1000;
	e.cbox.w=0;
	if(i==maxentoids)
		maxentoids++;
}

void AddPlatform(Sprite& sprite, int x, int y)
{
	for(int i=0;i<maxplatforms;i++)
		if(!platforms[i].exists)
			break;
	Platform& p=platforms[i];
	p.exists=true;
	p.sprite=&sprite;
	p.x=x;
	p.y=y;
	p.w=sprite.width;
	p.h=sprite.height;
	if(i==maxplatforms)
		maxplatforms++;
}

#include "entoids.h"

void ResetGame()
{
	global_win=0;
	global_lose=0;
	global_score=0;
	maxplatforms=0;
	maxentoids=0;
	scroll_x=0;
	scroll_y=0;
	AddEntoid(SE_Player, 160-6, 160);
	AddPlatform(spr_platform[0], 160-6-spr_platform[0].width/2, 180);
	AddEntoid(SE_Calvin, 230, 240);
	AddEntoid(SE_Drone, 64, 240);
}

void main()
{
	srand(time(NULL));

	if(!ptc_open("spatzap", WIDTH, HEIGHT))
		return;

	bool success=true;
	
	// load gfx
	if(!LoadSprite(spr_room[0], "data\\room1.png")) success=false;
	if(!LoadSprite(spr_room[1], "data\\room2.png")) success=false;
	if(!LoadSprite(spr_room[2], "data\\room3.png")) success=false;
	if(!LoadSprite(spr_room[3], "data\\room4.png")) success=false;
	if(!LoadSprite(spr_room[4], "data\\room5.png")) success=false;
	if(!LoadSprite(spr_room[5], "data\\roomjump.png")) success=false;
	if(!LoadSprite(spr_gem[0], "data\\gem1.png")) success=false;
	if(!LoadSprite(spr_gem[1], "data\\gem2.png")) success=false;
	if(!LoadSprite(spr_gem[2], "data\\gem3.png")) success=false;
	if(!LoadSprite(spr_gem[3], "data\\gem4.png")) success=false;
	if(!LoadSprite(spr_gem[4], "data\\gem5.png")) success=false;
	if(!LoadSprite(spr_spark[0], "data\\spark1.png")) success=false;
	if(!LoadSprite(spr_spark[1], "data\\spark2.png")) success=false;
	if(!LoadSprite(spr_spark[2], "data\\spark3.png")) success=false;
	if(!LoadSprite(spr_calvin[0], "data\\calvin1.png")) success=false;
	if(!LoadSprite(spr_calvin[1], "data\\calvin2.png")) success=false;
	if(!LoadSprite(spr_calvin[2], "data\\calvin3.png")) success=false;
	if(!LoadSprite(spr_drone[0], "data\\drone1.png")) success=false;
	if(!LoadSprite(spr_drone[1], "data\\drone2.png")) success=false;
	if(!LoadSprite(spr_enemy[0], "data\\enemy5.png")) success=false;
	if(!LoadSprite(spr_enemy[1], "data\\enemy6.png")) success=false;
	if(!LoadSprite(spr_enemy[2], "data\\enemy1.png")) success=false;
	if(!LoadSprite(spr_enemy[3], "data\\enemy2.png")) success=false;
	if(!LoadSprite(spr_enemy[4], "data\\enemy3.png")) success=false;
	if(!LoadSprite(spr_enemy[5], "data\\enemy4.png")) success=false;
	if(!LoadSprite(spr_platform[0], "data\\cliff1.png")) success=false; numptypes++;
	if(!LoadSprite(spr_platform[1], "data\\tbar1.png")) success=false; numptypes++;
	if(!LoadSprite(spr_platform[2], "data\\tbar2.png")) success=false; numptypes++;
	if(!LoadSprite(spr_platform[3], "data\\tbar1b.png")) success=false; numptypes++;
	if(!LoadSprite(spr_platform[4], "data\\tbox.png")) success=false; numptypes++;
	if(!LoadSprite(spr_border, "data\\border.png")) success=false;
	if(!LoadSprite(spr_bg1, "data\\bg1.png")) success=false;
	if(!LoadSprite(spr_bg2, "data\\bg2.png")) success=false;
	if(!LoadSprite(spr_win, "data\\win.png")) success=false;
	if(!LoadSprite(spr_lose, "data\\lose.png")) success=false;
	if(!success)
	{
		MessageBox(NULL, "Error loading graphics data", "Error", 0);
		playmu_close();
		ptc_close();
		ExitProcess(0);
	}

	SetEnvironmentVariable("PA_MIN_LATENCY_MSEC", "75");

	playmu_init();
	playmu_mastervolume(100);

	// load sounds
	if(!playmu_loadsound(sfx_jump, "data\\sfx\\jump.wav", 1)) success=false;
	if(!playmu_loadsound(sfx_gem, "data\\sfx\\gem.wav", 1)) success=false;
	if(!playmu_loadsound(sfx_win, "data\\sfx\\win.wav", 1)) success=false;
	if(!playmu_loadsound(sfx_lose, "data\\sfx\\lose.wav", 1)) success=false;
	if(!playmu_loadsound(sfx_splat, "data\\sfx\\splat.wav", 1)) success=false;
	if(!playmu_loadsound(sfx_splash, "data\\sfx\\splash.wav", 1)) success=false;
	if(!playmu_loadsound(sfx_build, "data\\sfx\\build.wav", 1)) success=false;
	if(!playmu_loadsound(sfx_drone, "data\\sfx\\drone.wav", 1)) success=false;
	if(!playmu_loadsong("data\\music\\bgm1.smu")) success=false;
	if(!success)
	{
		MessageBox(NULL, "Error loading sound data", "Error", 0);
		playmu_close();
		ptc_close();
		ExitProcess(0);
	}

	playmu_playsong(0);
	playmu_songvolume(50);

	AddEntoid(SE_Player, 160-6, 160);
	AddPlatform(spr_platform[0], 160-6-spr_platform[0].width/2, 180);

	AddEntoid(SE_Calvin, 230, 240);

//	AddEntoid(SE_Drone, -320*2+50+rnd(320*5-100), 240);
	AddEntoid(SE_Drone, 64, 240);

//	AddEntoid(SE_Enemy, 135, 240);

//	int p3dy=1;

	int n=0;

	for(;;)
	{
		if(GetAsyncKeyState(VK_ESCAPE)&KF_UP)
			break;

		n++;

		if(rnd(200)==0)
			AddEntoid(SE_Enemy, -320*2+50+rnd(320*5-100), 240);
/*
		platforms[2].y+=p3dy;
		if(platforms[2].y<30)
			p3dy=1;
		if(platforms[2].y>210)
			p3dy=-1;
*/

		if(global_win<6 && global_lose<6)
		{
			// update entoids
			for(int i=0;i<maxentoids;i++)
				if(entoids[i].exists)
					if(!entoids[i].TickFunc(entoids[i]))
						entoids[i].exists=false;
			// update entoid parameters ("instantly" for all entoids, since it's outside the update loop)
			int lastexisting=0;
			for(int i=0;i<maxentoids;i++)
				if(entoids[i].exists)
				{
					lastexisting=i;
					Entoid& e=entoids[i];
					// make new messages readable
					e.msgcount=e.msgncount;
					for(int mi=0;mi<e.msgcount;mi++)
						e.msg[mi]=e.msgn[mi];
					e.msgncount=0;
					// global box, for entoid collisions
					e.cbox=e.box;
					e.cbox.x+=e.x-e.box.x;
					e.cbox.y+=e.y-e.box.y;
				}
			maxentoids=lastexisting+1; // trim array size from empty spots at the end
		}

		// scroll, follow player
		Entoid& player=entoids[0];
		if(player.x+6<scroll_x+100) scroll_x=player.x+6-100;
		if(player.x+6>scroll_x+220) scroll_x=player.x+6-220;
		if(player.y+10<scroll_y+100) scroll_y=player.y+10-100;
		if(player.y+10>scroll_y+140) scroll_y=player.y+10-140;
		if(scroll_x<-320*2-16) scroll_x=-320*2-16;
		if(scroll_x>320*2+16) scroll_x=320*2+16;
		if(scroll_y<-240*4) scroll_y=-240*4;
		if(scroll_y>0) scroll_y=0;

		// ---------

		// draw background layers (clouds)
		for(int y=0;y<240;y++)
		{
			int si=((y+2000+scroll_y/2)&511)*512;
			int di=y*320;
			for(int x=2000+scroll_x/2;x<2320+scroll_x/2;x++)
				screen[di++]=spr_bg2.data[si+(x&511)];
		}
		for(int y=0;y<240;y++)
		{
			int si=((y+2000+scroll_y)&511)*512;
			int di=y*320;
			for(int x=2000+scroll_x;x<2320+scroll_x;x++)
			{
				DWORD color=spr_bg1.data[si+(x&511)];
				if(color!=0x00FFF0C8)
					screen[di++]=color;
				else
					di++;
			}
		}

		// draw platforms
		for(int i=0;i<maxplatforms;i++)
			if(platforms[i].exists)
				DrawSprite(*platforms[i].sprite, platforms[i].x-scroll_x, platforms[i].y-scroll_y);

		// draw entoids, one layer at a time
		for(int li=0;li<3;li++)
			for(int i=0;i<maxentoids;i++)
				if(entoids[i].exists && entoids[i].sprite!=NULL && entoids[i].layer==li)
				{
					Entoid& e=entoids[i];
					DrawSprite(*e.sprite, e.x-e.box.x-scroll_x, e.y-e.box.y-scroll_y, e.flipped);
				}

		// draw borders
		for(int i=-240*4;i<240;i+=16)
		{
			DrawSprite(spr_border, -320*2-16-scroll_x, i-scroll_y);
			DrawSprite(spr_border, 320*3-scroll_x, i-scroll_y);
		}

		int sty=200-scroll_y;
		for(int y=0;y<40;y++)
		{
			int dy=sty+y;
			if(dy>=240) break;
			int di=dy*320;
			int si=(sty-y)*320;
			int red=0x30-y*4;
			if(red<0) red=0;
			int green=0x50-y*3;
			if(green<0) green=0;
			int blue=0x40-y*3;
			if(blue<0) blue=0;
			DWORD wgradient=(red<<16)|(green<<8)|blue;
			if(y==0) // surface
			{
				for(int x=0;x<320;x++)
				{
					float v=sin((float)(n-x)*0.02f)*1.3f+sin((float)(n+x*2)*0.47f)+sin((float)(n-x*3)*0.12f)*1.5f+sin((float)(n+x*5)*0.15f)*0.9f;
					if(v>1.7f)
						screen[di++]=0xFFFFFF;
					else if(v>0.0f)
						screen[di++]=0xC0D0C0;
					else
						screen[di++]=0x90B090;
				}
			}
			else
			{
				for(int x=0;x<320;x++)
				{
					DWORD mirror=screen[si++];
					screen[di++]=((mirror>>2)&0x3F3F3F)+wgradient; // blend mirrored image with water gradient
				}
			}
		}
		
		// draw gui
		DrawBar(160-100, 10, 202, 6, 0x000000);
		DrawBar(160-100+1, 11, 200, 2, 0x605050);
		DrawBar(160-100+1, 13, 200, 2, 0x302050);
		DrawBar(160-100+1, 11, global_score*2, 4, 0xE08040);
		DrawBar(160-100+1, 12, global_score*2, 1, 0xF0D070);

		if(global_win>0)
		{
			if(global_win==1)
			{
//				playmu_stopmusic();
				playmu_playsound(sfx_win, 70);
			}
			global_win++;
			DrawSprite(spr_win, 160-spr_win.width/2, 120-spr_win.height/2);
			if(GetAsyncKeyState(VK_RETURN)&KF_UP)
				ResetGame();
		}
		if(global_lose>0)
		{
			if(global_lose==1)
			{
//				playmu_stopmusic();
				playmu_playsound(sfx_lose, 70);
			}
			global_lose++;
			DrawSprite(spr_lose, 160-spr_lose.width/2, 120-spr_lose.height/2);
			if(GetAsyncKeyState(VK_RETURN)&KF_UP)
				ResetGame();
		}

		// draw boxes (debug)
/*		if(n&1)
			for(int i=0;i<maxentoids;i++)
				if(entoids[i].exists)
				{
					Entoid& e=entoids[i];
					DrawBar(e.cbox.x, e.cbox.y, e.cbox.w, e.cbox.h, 0x8888FF);
				}
*/
		UpdateScreen();		

		Sleep(10);
	}

	playmu_fadeout(500);
	Sleep(500);

	GCFree();

	playmu_close();

	ptc_close();
	ExitProcess(0);
}

